﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using SlimDX;
using SlimDX.DirectInput;

namespace FDK
{
	public class CInputKeyboard : IInputDevice, IDisposable
	{
		public CInputKeyboard( IntPtr hWnd, DirectInput directInput )
		{
			this.e入力デバイス種別 = E入力デバイス種別.Keyboard;
			this.GUID = null;
			this.ID = 0;
			this.strデバイス名 = "Keyboard";

			#region [ キーボードデバイスを生成する。]
			//-----------------
			try
			{
				this.devKeyboard = new Keyboard( directInput );
				this.devKeyboard.SetCooperativeLevel( hWnd, CooperativeLevel.NoWinKey | CooperativeLevel.Foreground | CooperativeLevel.Nonexclusive );
				this.devKeyboard.Properties.BufferSize = 32;
			}
			catch
			{
				C共通.tDisposeする( ref this.devKeyboard );
				throw;
			}
			//-----------------
			#endregion
			#region [ キーボードデバイスを Acquire する。]
			//-----------------
			try
			{
				this.devKeyboard.Acquire();
			}
			catch( DirectInputException )
			{
				// 失敗しても、特に何もしない。
			}
			//-----------------
			#endregion

			#region [ キー状態配列を初期化。]
			//-----------------
			for( int i = 0; i < this.bKeyState.Length; i++ )
				this.bKeyState[ i ] = false;
			//-----------------
			#endregion
			#region [ タイマーを生成。]
			//-----------------
			this.timer = new CTimer( CTimer.E種別.MultiMedia );
			//-----------------
			#endregion
		}

		#region [ IInputDevice 実装 ]
		//-----------------
		public E入力デバイス種別 e入力デバイス種別
		{
			get;
			protected set;
		}
		public string GUID
		{
			get;
			protected set;
		}
		public int ID
		{
			get;
			protected set;
		}
		public List<CInputEvent> list入力イベント 
		{
			get;
			protected set;
		}
		public string strデバイス名
		{
			get;
			protected set;
		}

		public void tポーリング( bool bWindowがアクティブ中, bool bバッファ入力を使用する )
		{
			// 状態バッファをクリア。

			for( int i = 0; i < 256; i++ )
			{
				this.bKeyPushDown[ i ] = false;
				this.bKeyPullUp[ i ] = false;
			}


			// 入力イベントリストを作成。

			if( bWindowがアクティブ中 && 
				this.devKeyboard != null &&
				this.devKeyboard.Acquire().IsSuccess && 
				this.devKeyboard.Poll().IsSuccess )
			{
				this.list入力イベント = new List<CInputEvent>( 32 );

				if( bバッファ入力を使用する )
				{
					#region [ (A) バッファ入力 ]
					//-----------------------------
					var bufferedData = this.devKeyboard.GetBufferedData();

					if( Result.Last.IsFailure || bufferedData == null )
						return;

					foreach( KeyboardState data in bufferedData )
					{
						foreach( Key key in data.PressedKeys )
						{
							this.list入力イベント.Add(
								new CInputEvent() {
									GUID = this.GUID,
									ID = this.ID,
									nKey = (int) key,
									b押された = true,
									nTimeStamp = data.TimeStamp,
									nVelocity = CInput管理.n通常音量,
								} );

							this.bKeyState[ (int) key ] = true;
							this.bKeyPushDown[ (int) key ] = true;
						}
						foreach( Key key in data.ReleasedKeys )
						{
							this.list入力イベント.Add(
								new CInputEvent() {
									GUID = this.GUID,
									ID = this.ID,
									nKey = (int) key,
									b押された = false,
									nTimeStamp = data.TimeStamp,
									nVelocity = 0,
								} );

							this.bKeyState[ (int) key ] = false;
							this.bKeyPullUp[ (int) key ] = true;
						}
					}
					//-----------------------------
					#endregion
				}
				else
				{
					#region [ (B) 状態入力 ]
					//-----------------------------
					KeyboardState currentState = this.devKeyboard.GetCurrentState();

					if( Result.Last.IsFailure || currentState == null )
						return;

					foreach( Key key in currentState.PressedKeys )
					{
						if( this.bKeyState[ (int) key ] == false )
						{
							this.list入力イベント.Add(
								new CInputEvent() {
									GUID = this.GUID,
									ID = this.ID,
									nKey = (int) key,
									b押された = true,
									nTimeStamp = this.timer.nシステム時刻ms,
									nVelocity = CInput管理.n通常音量,
								} );

							this.bKeyState[ (int) key ] = true;
							this.bKeyPushDown[ (int) key ] = true;
						}
					}
					foreach( Key key in currentState.ReleasedKeys )
					{
						if( this.bKeyState[ (int) key ] == true )
						{
							this.list入力イベント.Add(
								new CInputEvent() {
									GUID = this.GUID,
									ID = this.ID,
									nKey = (int) key,
									b押された = false,
									nTimeStamp = this.timer.nシステム時刻ms,
									nVelocity = 0,
								} );

							this.bKeyState[ (int) key ] = false;
							this.bKeyPullUp[ (int) key ] = true;
						}
					}
					//-----------------------------
					#endregion
				}
			}
		}
		public bool bキーが押された( int nKey )
		{
			return this.bKeyPushDown[ nKey ];
		}
		public bool bキーが押された( Key key )
		{
			return this.bキーが押された( (int) key );
		}
		public bool bキーが押されている( int nKey )
		{
			return this.bKeyState[ nKey ];
		}
		public bool bキーが押されている( Key key )
		{
			return this.bキーが押されている( (int) key );
		}
		public bool bキーが離された( int nKey )
		{
			return this.bKeyPullUp[ nKey ];
		}
		public bool bキーが離された( Key key )
		{
			return this.bキーが離された( (int) key );
		}
		public bool bキーが離されている( int nKey )
		{
			return !this.bKeyState[ nKey ];
		}
		public bool bキーが離されている( Key key )
		{
			return this.bキーが離されている( (int) key );
		}
		//-----------------
		#endregion

		#region [ IDisposable 実装 ]
		//-----------------
		public void Dispose()
		{
			C共通.tDisposeする( ref this.devKeyboard );
			C共通.tDisposeする( ref this.timer );
		}
		//-----------------
		#endregion

		#region [ protected ]
		//-----------------
		protected bool[] bKeyPullUp = new bool[ 0x100 ];
		protected bool[] bKeyPushDown = new bool[ 0x100 ];
		protected bool[] bKeyState = new bool[ 0x100 ];
		protected Keyboard devKeyboard;
		protected CTimer timer;
		//-----------------
		#endregion
	}
}
