﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using SlimDX;
using SlimDX.DirectInput;

namespace FDK
{
	public class CInputMouse : IInputDevice, IDisposable
	{
		public const int nマウスの最大ボタン数 = 8;

		public CInputMouse( IntPtr hWnd, DirectInput directInput )
		{
			this.e入力デバイス種別 = E入力デバイス種別.Mouse;
			this.GUID = null;
			this.ID = 0;

			#region [ マウスデバイスを生成する。 ]
			//-----------------
			try
			{
				this.devMouse = new Mouse( directInput );
				this.devMouse.SetCooperativeLevel( hWnd, CooperativeLevel.Foreground | CooperativeLevel.Nonexclusive );
				this.devMouse.Properties.BufferSize = 32;
				this.strデバイス名 = this.devMouse.Information.ProductName;
			}
			catch
			{
				C共通.tDisposeする( ref this.devMouse );
				throw;
			}
			//-----------------
			#endregion
			#region [ マウスデバイスを Acquire する。]
			//-----------------
			try
			{
				this.devMouse.Acquire();
			}
			catch( DirectInputException )
			{
				// 失敗しても、特に何もしない。
			}
			//-----------------
			#endregion

			#region [ ボタン状態配列を初期化。]
			//-----------------
			for( int i = 0; i < this.bMouseState.Length; i++ )
				this.bMouseState[ i ] = false;
			//-----------------
			#endregion
			#region [ タイマーを生成。]
			//-----------------
			this.timer = new CTimer( CTimer.E種別.MultiMedia );
			//-----------------
			#endregion
		}

		#region [ IInputDevice 実装 ]
		//-----------------
		public E入力デバイス種別 e入力デバイス種別
		{
			get;
			protected set; 
		}
		public string GUID { 
			get; 
			protected set;
		}
		public int ID 
		{ 
			get;
			protected set;
		}
		public List<CInputEvent> list入力イベント 
		{
			get;
			protected set;
		}
		public string strデバイス名
		{
			get;
			protected set;
		}
		
		public void tポーリング( bool bWindowがアクティブ中, bool bバッファ入力を使用する )
		{
			for( int i = 0; i < 8; i++ )
			{
				this.bMousePushDown[ i ] = false;
				this.bMousePullUp[ i ] = false;
			}

			if( bWindowがアクティブ中 && 
				this.devMouse != null && 
				this.devMouse.Acquire().IsSuccess &&
				this.devMouse.Poll().IsSuccess )
			{
				this.list入力イベント = new List<CInputEvent>( 32 );

				if( bバッファ入力を使用する )
				{
					#region [ (A) バッファ入力 ]
					//-----------------------------
					var bufferedData = this.devMouse.GetBufferedData();

					if( Result.Last.IsFailure || bufferedData == null )
						return;

					foreach( MouseState data in bufferedData )
					{
						for( int button = 0; button < 8; button++ )
						{
							if( data.IsPressed( button ) )
							{
								this.list入力イベント.Add(
									new CInputEvent() {
										GUID = this.GUID,
										ID = this.ID,
										nKey = button,
										b押された = true,
										nTimeStamp = data.TimeStamp,
										nVelocity = CInput管理.n通常音量,
									} );

								this.bMouseState[ button ] = true;
								this.bMousePushDown[ button ] = true;
							}
							else if( data.IsReleased( button ) )
							{
								this.list入力イベント.Add(
									new CInputEvent() {
										GUID = this.GUID,
										ID = this.ID,
										nKey = button,
										b押された = false,
										nTimeStamp = data.TimeStamp,
										nVelocity = 0,
									} );

								this.bMouseState[ button ] = false;
								this.bMousePullUp[ button ] = true;
							}
						}
					}
					//-----------------------------
					#endregion
				}
				else
				{
					#region [ (B) 状態入力 ]
					//-----------------------------
					MouseState currentState = this.devMouse.GetCurrentState();

					if( Result.Last.IsFailure || currentState == null )
						return;

					bool[] buttons = currentState.GetButtons();

					for( int b = 0; ( b < buttons.Length ) && ( b < 8 ); b++ )
					{
						if( this.bMouseState[ b ] == false && buttons[ b ] == true )
						{
							this.list入力イベント.Add(
								new CInputEvent() {
									GUID = this.GUID,
									ID = this.ID,
									nKey = b,
									b押された = true,
									nTimeStamp = this.timer.nシステム時刻ms,
									nVelocity = CInput管理.n通常音量,
								} );

							this.bMouseState[ b ] = true;
							this.bMousePushDown[ b ] = true;
						}
						else if( this.bMouseState[ b ] == true && buttons[ b ] == false )
						{
							this.list入力イベント.Add(
								new CInputEvent() {
									GUID = this.GUID,
									ID = this.ID,
									nKey = b,
									b押された = false,
									nTimeStamp = this.timer.nシステム時刻ms,
									nVelocity = 0,
								} );

							this.bMouseState[ b ] = false;
							this.bMousePullUp[ b ] = true;
						}
					}
					//-----------------------------
					#endregion
				}
			}
		}
		public bool bキーが押された( int nButton )
		{
			return ( ( ( 0 <= nButton ) && ( nButton < 8 ) ) && this.bMousePushDown[ nButton ] );
		}
		public bool bキーが押された( Key key )
		{
			return this.bキーが押された( (int) key );
		}
		public bool bキーが押されている( int nButton )
		{
			return ( ( ( 0 <= nButton ) && ( nButton < 8 ) ) && this.bMouseState[ nButton ] );
		}
		public bool bキーが押されている( Key key )
		{
			return this.bキーが押されている( (int) key );
		}
		public bool bキーが離された( int nButton )
		{
			return ( ( ( 0 <= nButton ) && ( nButton < 8 ) ) && this.bMousePullUp[ nButton ] );
		}
		public bool bキーが離された( Key key )
		{
			return this.bキーが離された( (int) key );
		}
		public bool bキーが離されている( int nButton )
		{
			return ( ( ( 0 <= nButton ) && ( nButton < 8 ) ) && !this.bMouseState[ nButton ] );
		}
		public bool bキーが離されている( Key key )
		{
			return this.bキーが離されている( (int) key );
		}
		//-----------------
		#endregion

		#region [ IDisposable 実装 ]
		//-----------------
		public void Dispose()
		{
			C共通.tDisposeする( ref this.devMouse );
			C共通.tDisposeする( ref this.timer );
		}
		//-----------------
		#endregion

		#region [ protected ]
		//-----------------
		protected bool[] bMousePullUp = new bool[ 8 ];
		protected bool[] bMousePushDown = new bool[ 8 ];
		protected bool[] bMouseState = new bool[ 8 ];
		protected Mouse devMouse;
		protected CTimer timer;
		//-----------------
		#endregion
	}
}
