﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using System.Runtime.InteropServices;
using SlimDX;
using SlimDX.DirectInput;

namespace FDK
{
	public class CInput管理 : IDisposable
	{
		// 定数

		public static int n通常音量 = 110;


		// プロパティ

		public List<IInputDevice> list入力デバイス 
		{
			get;
			protected set;
		}
		public IInputDevice Keyboard
		{
			get
			{
				if( this._Keyboard != null )
				{
					return this._Keyboard;
				}
				foreach( IInputDevice device in this.list入力デバイス )
				{
					if( device.e入力デバイス種別 == E入力デバイス種別.Keyboard )
					{
						this._Keyboard = device;
						return device;
					}
				}
				return null;
			}
		}
		public IInputDevice Mouse
		{
			get
			{
				if( this._Mouse != null )
				{
					return this._Mouse;
				}
				foreach( IInputDevice device in this.list入力デバイス )
				{
					if( device.e入力デバイス種別 == E入力デバイス種別.Mouse )
					{
						this._Mouse = device;
						return device;
					}
				}
				return null;
			}
		}
		public int nMidiIn数
		{
			get;
			protected set;
		}
		
		/// <summary>どのウィンドウが今入力管理を行っているかは、このプロパティで確認できる。</summary>
		public IntPtr hWnd
		{
			get;
			protected set;
		}


		// コンストラクタ

		public CInput管理( IntPtr hWnd )
		{
			this.hWnd = hWnd;
			this.directInput = new DirectInput();
			this.timer = new CTimer( CTimer.E種別.MultiMedia );
			this.list入力デバイス = new List<IInputDevice>( 10 );


			// キーボード

			this.list入力デバイス.Add( new CInputKeyboard( hWnd, directInput ) );
			

			// マウス

			this.list入力デバイス.Add( new CInputMouse( hWnd, directInput ) );
			

			// ジョイパッド

			foreach( DeviceInstance instance in this.directInput.GetDevices( DeviceClass.GameController, DeviceEnumerationFlags.AttachedOnly ) )
				this.list入力デバイス.Add( new CInputJoystick( hWnd, instance, directInput ) );


			// MIDI In 機器

			this.proc = new CWin32.MidiInProc( this.MidiInCallback );
			this.nMidiIn数 = (int) CWin32.midiInGetNumDevs();

			for( uint i = 0; i < this.nMidiIn数; i++ )
			{
				var item = new CInputMIDI( i );
				this.list入力デバイス.Add( item );

				var lpMidiInCaps = new CWin32.MIDIINCAPS();
				uint ret = CWin32.midiInGetDevCaps( i, ref lpMidiInCaps, (uint) Marshal.SizeOf( lpMidiInCaps ) );
				if( ret != 0 )
					continue;	// エラー

				item.strデバイス名 = lpMidiInCaps.szPname;

				if( ( CWin32.midiInOpen( ref item.hMidiIn, i, this.proc, 0, 0x30000 ) == 0 ) && ( item.hMidiIn != 0 ) )
				{
					// 入力受け付け開始
					CWin32.midiInStart( item.hMidiIn );
				}
				else
				{ 
					// 失敗
				}
			}
		}
		
		
		// メソッド

		public IInputDevice Joystick( int ID )
		{
			foreach( IInputDevice device in this.list入力デバイス )
			{
				if( ( device.e入力デバイス種別 == E入力デバイス種別.Joystick ) && ( device.ID == ID ) )
					return device;
			}
			return null;
		}
		public IInputDevice Joystick( string GUID )
		{
			foreach( IInputDevice device in this.list入力デバイス )
			{
				if( ( device.e入力デバイス種別 == E入力デバイス種別.Joystick ) && device.GUID.Equals( GUID ) )
					return device;
			}
			return null;
		}
		public IInputDevice MidiIn( int ID )
		{
			foreach( IInputDevice device in this.list入力デバイス )
			{
				if( ( device.e入力デバイス種別 == E入力デバイス種別.MidiIn ) && ( device.ID == ID ) )
					return device;
			}
			return null;
		}
		public void tポーリング( bool bWindowがアクティブ中, bool bバッファ入力を使用する )
		{
			lock( this.objMidiIn排他用 )
			{
				for( int i = this.list入力デバイス.Count - 1; i >= 0; i-- )
				{
					IInputDevice device = this.list入力デバイス[ i ];
					try
					{
						device.tポーリング( bWindowがアクティブ中, bバッファ入力を使用する );
					}
					catch( DirectInputException )
					{
						// ポーリングに失敗したので、このデバイスを除外する。

						C共通.tDisposeする( ref device );
						this.list入力デバイス.RemoveAt( i );
					}
				}
			}
		}

		#region [ IDisposable 実装 ]
		//-----------------
		public void Dispose()
		{
			if( this.list入力デバイス.Count == 0 )
				return;

			#region [ 全MIDI入力デバイスを停止する。 ]
			//-----------------
			foreach( IInputDevice device in this.list入力デバイス )
			{
				var devMIDI = device as CInputMIDI;
				if( devMIDI != null )
				{
					CWin32.midiInStop( devMIDI.hMidiIn );
					CWin32.midiInReset( devMIDI.hMidiIn );
					CWin32.midiInClose( devMIDI.hMidiIn );
				}
			}
			//-----------------
			#endregion
			#region [ 全入力デバイスを解放する。 ]
			//-----------------
			foreach( IInputDevice dev in this.list入力デバイス )
				dev.Dispose();
			//-----------------
			#endregion

			this.list入力デバイス.Clear();

			C共通.tDisposeする( ref this.directInput );
			C共通.tDisposeする( ref this.timer );
		}
		//-----------------
		#endregion

		#region [ protected ]
		//-----------------
		protected DirectInput directInput;
		protected IInputDevice _Keyboard;
		protected IInputDevice _Mouse;
		protected List<uint> listHMIDIIN = new List<uint>( 8 );
		protected object objMidiIn排他用 = new object();
		protected CWin32.MidiInProc proc;
		protected CTimer timer;

		protected void MidiInCallback( uint hMidiIn, uint wMsg, int dwInstance, int dwParam1, int dwParam2 )
		{
			int nStatus = dwParam1 & 0xF0;
			long time = this.timer.nシステム時刻ms;	// lock前に取得

			lock( this.objMidiIn排他用 )
			{
				if( ( this.list入力デバイス != null ) && ( this.list入力デバイス.Count != 0 ) )
				{
					foreach( IInputDevice device in this.list入力デバイス )
					{
						var tmidi = device as CInputMIDI;
						if( ( tmidi != null ) && ( tmidi.hMidiIn == hMidiIn ) )
						{
							tmidi.tメッセージからMIDI信号のみ受信( wMsg, dwInstance, dwParam1, dwParam2, time );
							break;
						}
					}
				}
			}
		}
		//-----------------
		#endregion
	}
}
