﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Drawing;
using System.Drawing.Imaging;
using System.IO;
using System.Diagnostics;
using SlimDX;
using SlimDX.Direct3D9;

namespace FDK
{
	public class CTexture : IDisposable
	{
		// プロパティ

		public bool b加算合成 = false;
		/// <summary>
		/// 透明:0～255:不透明。
		/// </summary>
		public int n透明度
		{
			get
			{
				return this._透明度;
			}
			set
			{
				this._透明度 = Math.Min( Math.Max( value, 0 ), 255 );
			}
		}
		public Size szテクスチャサイズ
		{
			get; 
			protected set;
		}
		public Size sz画像サイズ
		{
			get;
			protected set;
		}
		public Texture texture
		{
			get;
			protected set;
		}
		public Format Format
		{
			get;
			protected set;
		}
		public Vector3 vc拡大縮小倍率 = new Vector3( 1f, 1f, 1f );
		/// <summary>
		/// <para>描画範囲を示す矩形のリスト。</para>
		/// <para>本クラスでは使用せず、使用方法はすべてユーザに委ねる。未使用の場合は空でも可。nullチェックは不要。</para>
		/// </summary>
		public List<Rectangle> list描画範囲
		{
			get;
			protected set;
		}


		// 画面が変わるたび以下のプロパティを設定し治すこと。

		public static Size sz論理画面 = Size.Empty;
		public static Size sz物理画面 = Size.Empty;
		public static  Rectangle rc物理画面描画領域 = Rectangle.Empty;
		/// <summary>
		/// <para>論理画面を1とする場合の物理画面の倍率。</para>
		/// <para>論理値×画面比率＝物理値。</para>
		/// </summary>
		public static float f画面比率 = 1.0f;


		// コンストラクタ

		public CTexture()
		{
			this.sz画像サイズ = new Size( 0, 0 );
			this.szテクスチャサイズ = new Size( 0, 0 );
			this._透明度 = 0xff;
			this.texture = null;
			this.list描画範囲 = new List<Rectangle>();
		}
		
		/// <summary>
		/// <para>指定されたビットマップオブジェクトから Managed テクスチャを作成する。</para>
		/// <para>テクスチャのサイズは、BITMAP画像のサイズ以上、かつ、D3D9デバイスで生成可能な最小のサイズに自動的に調節される。
		/// その際、テクスチャの調節後のサイズにあわせた画像の拡大縮小は行わない。</para>
		/// <para>その他、ミップマップ数は 1、Usage は None、Pool は Managed、イメージフィルタは Point、ミップマップフィルタは
		/// None、カラーキーはなし（透過しない）になる。</para>
		/// </summary>
		/// <param name="device">Direct3D9 デバイス。</param>
		/// <param name="bitmap">作成元のビットマップ。</param>
		/// <param name="format">テクスチャのフォーマット。</param>
		/// <exception cref="CTextureCreateFailedException">テクスチャの作成に失敗しました。</exception>
		public CTexture( Device device, Bitmap bitmap, Format format )
			: this( device, bitmap, format, false )
		{ 
		}

		/// <summary>
		/// <para>指定されたビットマップオブジェクトから Managed テクスチャを作成する。</para>
		/// <para>テクスチャのサイズは、BITMAP画像のサイズ以上、かつ、D3D9デバイスで生成可能な最小のサイズに自動的に調節される。
		/// その際、テクスチャの調節後のサイズにあわせた画像の拡大縮小は行わない。</para>
		/// <para>その他、ミップマップ数は 1、Usage は None、Pool は Managed、イメージフィルタは Point、ミップマップフィルタは
		/// None、カラーキーは、（b黒を透過する=trueの場合）0xFF000000（完全なる黒を透過）になる。</para>
		/// </summary>
		/// <param name="device">Direct3D9 デバイス。</param>
		/// <param name="bitmap">作成元のビットマップ。</param>
		/// <param name="format">テクスチャのフォーマット。</param>
		/// <param name="b黒を透過する">黒を透過するなら true。</param>
		/// <exception cref="CTextureCreateFailedException">テクスチャの作成に失敗しました。</exception>
		public CTexture( Device device, Bitmap bitmap, Format format, bool b黒を透過する )
			: this()
		{
			try
			{
				this.Format = format;
				this.sz画像サイズ = new Size( bitmap.Width, bitmap.Height );
				this.szテクスチャサイズ = this.t指定されたサイズを超えない最適なテクスチャサイズを返す( device, this.sz画像サイズ );

				using( var stream = new MemoryStream() )
				{
					bitmap.Save( stream, ImageFormat.Bmp );
					stream.Seek( 0L, SeekOrigin.Begin );
					int colorKey = ( b黒を透過する ) ? unchecked( (int) 0xFF000000 ) : 0;
					this.texture = Texture.FromStream( device, stream, this.szテクスチャサイズ.Width, this.szテクスチャサイズ.Height, 1, Usage.None, format, Pool.Managed, Filter.Point, Filter.None, colorKey );
				}
			}
			catch
			{
				this.Dispose();
				throw;
			}

			#region [ テクスチャ解放漏れの調査用コード。 ]
			//-----------------
			//			Debug.Assert( n生存数 != 0 );			
			//			Debug.WriteLine( n生存数 + ": Bitmapからテクスチャ作成" ); n生存ID = n生存数++;
			//-----------------
			#endregion
		}

		/// <summary>
		/// <para>空の Managed テクスチャを作成する。</para>
		/// <para>テクスチャのサイズは、指定された希望サイズ以上、かつ、D3D9デバイスで生成可能な最小のサイズに自動的に調節される。
		/// その際、テクスチャの調節後のサイズにあわせた画像の拡大縮小は行わない。</para>
		/// <para>テクスチャのテクセルデータは未初期化。（おそらくゴミデータが入ったまま。）</para>
		/// <para>その他、ミップマップ数は 1、Usage は None、イメージフィルタは Point、ミップマップフィルタは None、
		/// カラーキーは 0x00000000（透過しない）になる。</para>
		/// </summary>
		/// <param name="device">Direct3D9 デバイス。</param>
		/// <param name="n幅">テクスチャの幅（希望値）。</param>
		/// <param name="n高さ">テクスチャの高さ（希望値）。</param>
		/// <param name="format">テクスチャのフォーマット。</param>
		/// <exception cref="CTextureCreateFailedException">テクスチャの作成に失敗しました。</exception>
		public CTexture( Device device, int n幅, int n高さ, Format format )
			: this( device, n幅, n高さ, format, Pool.Managed )
		{
		}
		
		/// <summary>
		/// <para>指定された画像ファイルから Managed テクスチャを作成する。</para>
		/// <para>利用可能な画像形式は、BMP, JPG, PNG, TGA, DDS, PPM, DIB, HDR, PFM のいずれか。</para>
		/// </summary>
		/// <param name="device">Direct3D9 デバイス。</param>
		/// <param name="strファイル名">画像ファイル名。</param>
		/// <param name="format">テクスチャのフォーマット。</param>
		/// <param name="b黒を透過する">画像の黒（0xFFFFFFFF）を透過させるなら true。</param>
		/// <exception cref="CTextureCreateFailedException">テクスチャの作成に失敗しました。</exception>
		public CTexture( Device device, string strファイル名, Format format, bool b黒を透過する )
			: this( device, strファイル名, format, b黒を透過する, Pool.Managed )
		{
		}
		
		/// <summary>
		/// <para>空のテクスチャを作成する。</para>
		/// <para>テクスチャのサイズは、指定された希望サイズ以上、かつ、D3D9デバイスで生成可能な最小のサイズに自動的に調節される。
		/// その際、テクスチャの調節後のサイズにあわせた画像の拡大縮小は行わない。</para>
		/// <para>テクスチャのテクセルデータは未初期化。（おそらくゴミデータが入ったまま。）</para>
		/// <para>その他、ミップマップ数は 1、Usage は None、イメージフィルタは Point、ミップマップフィルタは None、
		/// カラーキーは 0x00000000（透過しない）になる。</para>
		/// </summary>
		/// <param name="device">Direct3D9 デバイス。</param>
		/// <param name="n幅">テクスチャの幅（希望値）。</param>
		/// <param name="n高さ">テクスチャの高さ（希望値）。</param>
		/// <param name="format">テクスチャのフォーマット。</param>
		/// <param name="pool">テクスチャの管理方法。</param>
		/// <exception cref="CTextureCreateFailedException">テクスチャの作成に失敗しました。</exception>
		public CTexture( Device device, int n幅, int n高さ, Format format, Pool pool )
			: this()
		{
			try
			{
				this.Format = format;
				this.sz画像サイズ = new Size( n幅, n高さ );
				this.szテクスチャサイズ = this.t指定されたサイズを超えない最適なテクスチャサイズを返す( device, this.sz画像サイズ );

				using( Bitmap bitmap = new Bitmap( 1, 1 ) )
				{
					using( Graphics graphics = Graphics.FromImage( bitmap ) )
					{
						graphics.FillRectangle( Brushes.Black, 0, 0, 1, 1 );
					}

					using( MemoryStream stream = new MemoryStream() )
					{
						bitmap.Save( stream, ImageFormat.Bmp );
						stream.Seek( 0L, SeekOrigin.Begin );
						this.texture = Texture.FromStream( device, stream, n幅, n高さ, 1, Usage.None, format, pool, Filter.Point, Filter.None, 0 );
					}
				}
			}
			catch
			{
				this.Dispose();
				throw;
			}

			#region [ テクスチャ解放漏れの調査用コード。 ]
			//-----------------
//			Debug.Assert( n生存数 != 0 );
//			Debug.WriteLine( n生存数 + ": 幅・高さ・Format・Poolからテクスチャ作成" ); n生存ID = n生存数++;
			//-----------------
			#endregion
		}
		
		/// <summary>
		/// <para>画像ファイルからテクスチャを生成する。</para>
		/// <para>利用可能な画像形式は、BMP, JPG, PNG, TGA, DDS, PPM, DIB, HDR, PFM のいずれか。</para>
		/// <para>テクスチャのサイズは、画像のサイズ以上、かつ、D3D9デバイスで生成可能な最小のサイズに自動的に調節される。
		/// その際、テクスチャの調節後のサイズにあわせた画像の拡大縮小は行わない。</para>
		/// <para>その他、ミップマップ数は 1、Usage は None、イメージフィルタは Point、ミップマップフィルタは None になる。</para>
		/// </summary>
		/// <param name="device">Direct3D9 デバイス。</param>
		/// <param name="strファイル名">画像ファイル名。</param>
		/// <param name="format">テクスチャのフォーマット。</param>
		/// <param name="b黒を透過する">画像の黒（0xFFFFFFFF）を透過させるなら true。</param>
		/// <param name="pool">テクスチャの管理方法。</param>
		/// <exception cref="CTextureCreateFailedException">テクスチャの作成に失敗しました。</exception>
		public CTexture( Device device, string strファイル名, Format format, bool b黒を透過する, Pool pool )
			: this()
		{
			try
			{
				if( !File.Exists( strファイル名 ) )		// ImageInformation では FileNotFound 例外は返ってこないので、ここで自分でチェックする。わかりやすいログのために。
					throw new FileNotFoundException( string.Format( "ファイルが存在しません。\n[{0}]", strファイル名 ) );

				var information = ImageInformation.FromFile( strファイル名 );

				this.Format = format;
				this.sz画像サイズ = new Size( information.Width, information.Height );
				this.szテクスチャサイズ = this.t指定されたサイズを超えない最適なテクスチャサイズを返す( device, this.sz画像サイズ );
				this.texture = Texture.FromFile( device, strファイル名, this.sz画像サイズ.Width, this.sz画像サイズ.Height, 1, Usage.None, format, pool, Filter.Point, Filter.None, b黒を透過する ? -16777216 : 0 );
			}
			catch
			{
				this.Dispose();
				throw;
			}

			#region [ テクスチャ解放漏れの調査用コード。 ]
			//-----------------
//			Debug.Assert( n生存数 != 0 );
//			Debug.WriteLine( n生存数 + ": ファイル[" + strファイル名 + "]・Pool からテクスチャ作成" ); n生存ID = n生存数++;
			//-----------------
			#endregion
		}


		// メソッド

		public static CTexture t文字列テクスチャを作成する( string str文字列, Device D3DDevice )
		{
			return CTexture.t文字列テクスチャを作成する( str文字列, D3DDevice, Format.A8R8G8B8 );
		}
		public static CTexture t文字列テクスチャを作成する( string str文字列, Device D3DDevice, Format TextureFormat )
		{
			CTexture tx文字列 = null;


			// 標準ゴシック・16px のフォントを使用。

			using( System.Drawing.Font ftフォント = new System.Drawing.Font( FontFamily.GenericSansSerif, 16.0f, FontStyle.Regular, GraphicsUnit.Pixel ) )
			{
				tx文字列 = CTexture.t文字列テクスチャを作成する( str文字列, ftフォント, D3DDevice, TextureFormat );
			}

			return tx文字列;
		}
		public static CTexture t文字列テクスチャを作成する( string str文字列, System.Drawing.Font ftフォント, Device D3DDevice, Format TextureFormat )
		{
			SizeF sz文字列のサイズ;

			#region [ 文字列のサイズ(SizeF)[px] を取得する。]
			//-----------------
			using( var b = new Bitmap( 1, 1 ) )
			using( Graphics g = Graphics.FromImage( b ) )
			{
				g.PageUnit = GraphicsUnit.Pixel;
				sz文字列のサイズ = g.MeasureString( str文字列, ftフォント );
			}
			//-----------------
			#endregion

			CTexture tx文字列;

			#region [ 取得したサイズのテクスチャを作成する。 ]
			//-----------------
			using( var bitmap = new Bitmap( (int) sz文字列のサイズ.Width, (int) sz文字列のサイズ.Height, PixelFormat.Format32bppArgb ) )
			using( var g = Graphics.FromImage( bitmap ) )
			{
				g.DrawString( str文字列, ftフォント, Brushes.White, 0.0f, 0.0f );
				tx文字列 = new CTexture( D3DDevice, bitmap, TextureFormat );
			}
			//-----------------
			#endregion

			return tx文字列;
		}

		public static Vector3 t論理画面座標をワールド座標へ変換する( int x, int y )
		{
			return CTexture.t論理画面座標をワールド座標へ変換する( new Vector3( (float) x, (float) y, 0f ) );
		}
		public static Vector3 t論理画面座標をワールド座標へ変換する( float x, float y )
		{
			return CTexture.t論理画面座標をワールド座標へ変換する( new Vector3( x, y, 0f ) );
		}
		public static Vector3 t論理画面座標をワールド座標へ変換する( Point pt論理画面座標 )
		{
			return CTexture.t論理画面座標をワールド座標へ変換する( new Vector3( pt論理画面座標.X, pt論理画面座標.Y, 0.0f ) );
		}
		public static Vector3 t論理画面座標をワールド座標へ変換する( Vector2 v2論理画面座標 )
		{
			return CTexture.t論理画面座標をワールド座標へ変換する( new Vector3( v2論理画面座標, 0f ) );
		}
		public static Vector3 t論理画面座標をワールド座標へ変換する( Vector3 v3論理画面座標 )
		{
			return new Vector3(
				( v3論理画面座標.X - ( CTexture.sz論理画面.Width / 2.0f ) ) * CTexture.f画面比率,
				( -( v3論理画面座標.Y - ( CTexture.sz論理画面.Height / 2.0f ) ) * CTexture.f画面比率 ),
				v3論理画面座標.Z );
		}

		// 論理画面座標で描画する。
		public void t2D描画( Device device, int x, int y )
		{
			this.t2D描画( device, x, y, new Rectangle( 0, 0, this.sz画像サイズ.Width, this.sz画像サイズ.Height ) );
		}
		public void t2D描画( Device device, int x, int y, Rectangle rc画像内の描画領域 )
		{
			this.t2D描画( device, x, y, 1f, rc画像内の描画領域 );
		}
		public void t2D描画( Device device, int x, int y, float depth, Rectangle rc画像内の描画領域 )
		{
			if( this.texture == null )
				throw new InvalidOperationException( "テクスチャは生成されていません。" );

			this.tレンダリングステートの設定( device );

			float fx = x * CTexture.f画面比率 + CTexture.rc物理画面描画領域.X - 0.5f;	// -0.5 は座標とピクセルの誤差を吸収するための座標補正値。(MSDN参照)
			float fy = y * CTexture.f画面比率 + CTexture.rc物理画面描画領域.Y - 0.5f;	//
			float w = rc画像内の描画領域.Width * this.vc拡大縮小倍率.X * CTexture.f画面比率;
			float h = rc画像内の描画領域.Height * this.vc拡大縮小倍率.Y * CTexture.f画面比率;
			float f左U値 = ( (float) rc画像内の描画領域.Left ) / ( (float) this.szテクスチャサイズ.Width );
			float f右U値 = ( (float) rc画像内の描画領域.Right ) / ( (float) this.szテクスチャサイズ.Width );
			float f上V値 = ( (float) rc画像内の描画領域.Top ) / ( (float) this.szテクスチャサイズ.Height );
			float f下V値 = ( (float) rc画像内の描画領域.Bottom ) / ( (float) this.szテクスチャサイズ.Height );
			int color = new Color4( ( (float) this._透明度 ) / 255f, 1f, 1f, 1f ).ToArgb();

			// 以下、マネージドオブジェクトの量産を抑えるため new は使わない。

			this.cvTransformedColoredVertexies[ 0 ].TextureCoordinates.X = f左U値;
			this.cvTransformedColoredVertexies[ 0 ].TextureCoordinates.Y = f上V値;
			this.cvTransformedColoredVertexies[ 0 ].Position.X = fx;
			this.cvTransformedColoredVertexies[ 0 ].Position.Y = fy;
			this.cvTransformedColoredVertexies[ 0 ].Position.Z = depth;
			this.cvTransformedColoredVertexies[ 0 ].Position.W = 1.0f;
			this.cvTransformedColoredVertexies[ 0 ].Color = color;

			this.cvTransformedColoredVertexies[ 1 ].TextureCoordinates.X = f右U値;
			this.cvTransformedColoredVertexies[ 1 ].TextureCoordinates.Y = f上V値;
			this.cvTransformedColoredVertexies[ 1 ].Position.X = fx + w;
			this.cvTransformedColoredVertexies[ 1 ].Position.Y = fy;
			this.cvTransformedColoredVertexies[ 1 ].Position.Z = depth;
			this.cvTransformedColoredVertexies[ 1 ].Position.W = 1.0f;
			this.cvTransformedColoredVertexies[ 1 ].Color = color;

			this.cvTransformedColoredVertexies[ 2 ].TextureCoordinates.X = f左U値;
			this.cvTransformedColoredVertexies[ 2 ].TextureCoordinates.Y = f下V値;
			this.cvTransformedColoredVertexies[ 2 ].Position.X = fx;
			this.cvTransformedColoredVertexies[ 2 ].Position.Y = fy + h;
			this.cvTransformedColoredVertexies[ 2 ].Position.Z = depth;
			this.cvTransformedColoredVertexies[ 2 ].Position.W = 1.0f;
			this.cvTransformedColoredVertexies[ 2 ].Color = color;

			this.cvTransformedColoredVertexies[ 3 ].TextureCoordinates.X = f右U値;
			this.cvTransformedColoredVertexies[ 3 ].TextureCoordinates.Y = f下V値;
			this.cvTransformedColoredVertexies[ 3 ].Position.X = fx + w;
			this.cvTransformedColoredVertexies[ 3 ].Position.Y = fy + h;
			this.cvTransformedColoredVertexies[ 3 ].Position.Z = depth;
			this.cvTransformedColoredVertexies[ 3 ].Position.W = 1.0f;
			this.cvTransformedColoredVertexies[ 3 ].Color = color;

			device.SetTexture( 0, this.texture );
			device.VertexFormat = TransformedColoredTexturedVertex.Format;
			device.DrawUserPrimitives( PrimitiveType.TriangleStrip, 2, this.cvTransformedColoredVertexies );
		}
		public void t2D描画( Device device, Point pt )
		{
			this.t2D描画( device, pt.X, pt.Y, 1f, new Rectangle( 0, 0, this.sz画像サイズ.Width, this.sz画像サイズ.Height ) );
		}
		public void t2D描画( Device device, Point pt, Rectangle rc画像内の描画領域 )
		{
			this.t2D描画( device, pt.X, pt.Y, 1f, rc画像内の描画領域 );
		}
		public void t2D描画( Device device, Point pt, float depth, Rectangle rc画像内の描画領域 )
		{
			this.t2D描画( device, pt.X, pt.Y, depth, rc画像内の描画領域 );
		}

		// 物理画面座標で描画する。
		public void t3D描画( Device device, Matrix mat )
		{
			this.t3D描画( device, mat, new Rectangle( 0, 0, this.sz画像サイズ.Width, this.sz画像サイズ.Height ) );
		}
		public void t3D描画( Device device, Matrix mat, Rectangle rc画像内の描画領域 )
		{
			if( this.texture == null )
				return;

			float x = ( (float) rc画像内の描画領域.Width ) / 2.0f;
			float y = ( (float) rc画像内の描画領域.Height ) / 2.0f;
			float z = 0.0f;
			float f左U値 = ( (float) rc画像内の描画領域.Left ) / ( (float) this.szテクスチャサイズ.Width );
			float f右U値 = ( (float) rc画像内の描画領域.Right ) / ( (float) this.szテクスチャサイズ.Width );
			float f上V値 = ( (float) rc画像内の描画領域.Top ) / ( (float) this.szテクスチャサイズ.Height );
			float f下V値 = ( (float) rc画像内の描画領域.Bottom ) / ( (float) this.szテクスチャサイズ.Height );
			int color = new Color4( ( (float) this._透明度 ) / 255.0f, 1.0f, 1.0f, 1.0f ).ToArgb();

			// 以下、マネージドオブジェクトの量産を抑えるため new は使わない。

			this.cvPositionColoredVertexies[ 0 ].TextureCoordinates.X = f左U値;
			this.cvPositionColoredVertexies[ 0 ].TextureCoordinates.Y = f上V値;
			this.cvPositionColoredVertexies[ 0 ].Position.X = -x;
			this.cvPositionColoredVertexies[ 0 ].Position.Y = y;
			this.cvPositionColoredVertexies[ 0 ].Position.Z = z;
			this.cvPositionColoredVertexies[ 0 ].Color = color;

			this.cvPositionColoredVertexies[ 1 ].TextureCoordinates.X = f右U値;
			this.cvPositionColoredVertexies[ 1 ].TextureCoordinates.Y = f上V値;
			this.cvPositionColoredVertexies[ 1 ].Position.X = x;
			this.cvPositionColoredVertexies[ 1 ].Position.Y = y;
			this.cvPositionColoredVertexies[ 1 ].Position.Z = z;
			this.cvPositionColoredVertexies[ 1 ].Color = color;

			this.cvPositionColoredVertexies[ 2 ].TextureCoordinates.X = f左U値;
			this.cvPositionColoredVertexies[ 2 ].TextureCoordinates.Y = f下V値;
			this.cvPositionColoredVertexies[ 2 ].Position.X = -x;
			this.cvPositionColoredVertexies[ 2 ].Position.Y = -y;
			this.cvPositionColoredVertexies[ 2 ].Position.Z = z;
			this.cvPositionColoredVertexies[ 2 ].Color = color;

			this.cvPositionColoredVertexies[ 3 ].TextureCoordinates.X = f右U値;
			this.cvPositionColoredVertexies[ 3 ].TextureCoordinates.Y = f下V値;
			this.cvPositionColoredVertexies[ 3 ].Position.X = x;
			this.cvPositionColoredVertexies[ 3 ].Position.Y = -y;
			this.cvPositionColoredVertexies[ 3 ].Position.Z = z;
			this.cvPositionColoredVertexies[ 3 ].Color = color;

			this.tレンダリングステートの設定( device );
			device.SetTransform( TransformState.World, mat );
			device.SetTexture( 0, this.texture );
			device.VertexFormat = PositionColoredTexturedVertex.Format;
			device.DrawUserPrimitives( PrimitiveType.TriangleStrip, 4, this.cvPositionColoredVertexies );
		}

		// 論理画面を無視し、物理画面の座標により描画する。
		public void t物理2D描画( Device device, int x, int y )
		{
			this.t物理2D描画( device, x, y, new Rectangle( 0, 0, this.sz画像サイズ.Width, this.sz画像サイズ.Height ) );
		}
		public void t物理2D描画( Device device, int x, int y, Rectangle rc画像内の描画領域 )
		{
			this.t物理2D描画( device, x, y, 1f, rc画像内の描画領域 );
		}
		public void t物理2D描画( Device device, int x, int y, float depth, Rectangle rc画像内の描画領域 )
		{
			if( this.texture == null )
				throw new InvalidOperationException( "テクスチャは生成されていません。" );

			this.tレンダリングステートの設定( device );

			float fx = x - 0.5f;	// -0.5 は座標とピクセルの誤差を吸収するための座標補正値。(MSDN参照)
			float fy = y - 0.5f;	//
			float w = rc画像内の描画領域.Width * this.vc拡大縮小倍率.X;
			float h = rc画像内の描画領域.Height * this.vc拡大縮小倍率.Y;
			float f左U値 = ( (float) rc画像内の描画領域.Left ) / ( (float) this.szテクスチャサイズ.Width );
			float f右U値 = ( (float) rc画像内の描画領域.Right ) / ( (float) this.szテクスチャサイズ.Width );
			float f上V値 = ( (float) rc画像内の描画領域.Top ) / ( (float) this.szテクスチャサイズ.Height );
			float f下V値 = ( (float) rc画像内の描画領域.Bottom ) / ( (float) this.szテクスチャサイズ.Height );
			int color = new Color4( ( (float) this._透明度 ) / 255f, 1f, 1f, 1f ).ToArgb();

			// 以下、マネージドオブジェクトの量産を抑えるため new は使わない。

			this.cvTransformedColoredVertexies[ 0 ].TextureCoordinates.X = f左U値;
			this.cvTransformedColoredVertexies[ 0 ].TextureCoordinates.Y = f上V値;
			this.cvTransformedColoredVertexies[ 0 ].Position.X = fx;
			this.cvTransformedColoredVertexies[ 0 ].Position.Y = fy;
			this.cvTransformedColoredVertexies[ 0 ].Position.Z = depth;
			this.cvTransformedColoredVertexies[ 0 ].Position.W = 1.0f;
			this.cvTransformedColoredVertexies[ 0 ].Color = color;

			this.cvTransformedColoredVertexies[ 1 ].TextureCoordinates.X = f右U値;
			this.cvTransformedColoredVertexies[ 1 ].TextureCoordinates.Y = f上V値;
			this.cvTransformedColoredVertexies[ 1 ].Position.X = fx + w;
			this.cvTransformedColoredVertexies[ 1 ].Position.Y = fy;
			this.cvTransformedColoredVertexies[ 1 ].Position.Z = depth;
			this.cvTransformedColoredVertexies[ 1 ].Position.W = 1.0f;
			this.cvTransformedColoredVertexies[ 1 ].Color = color;

			this.cvTransformedColoredVertexies[ 2 ].TextureCoordinates.X = f左U値;
			this.cvTransformedColoredVertexies[ 2 ].TextureCoordinates.Y = f下V値;
			this.cvTransformedColoredVertexies[ 2 ].Position.X = fx;
			this.cvTransformedColoredVertexies[ 2 ].Position.Y = fy + h;
			this.cvTransformedColoredVertexies[ 2 ].Position.Z = depth;
			this.cvTransformedColoredVertexies[ 2 ].Position.W = 1.0f;
			this.cvTransformedColoredVertexies[ 2 ].Color = color;

			this.cvTransformedColoredVertexies[ 3 ].TextureCoordinates.X = f右U値;
			this.cvTransformedColoredVertexies[ 3 ].TextureCoordinates.Y = f下V値;
			this.cvTransformedColoredVertexies[ 3 ].Position.X = fx + w;
			this.cvTransformedColoredVertexies[ 3 ].Position.Y = fy + h;
			this.cvTransformedColoredVertexies[ 3 ].Position.Z = depth;
			this.cvTransformedColoredVertexies[ 3 ].Position.W = 1.0f;
			this.cvTransformedColoredVertexies[ 3 ].Color = color;

			device.SetTexture( 0, this.texture );
			device.VertexFormat = TransformedColoredTexturedVertex.Format;
			device.DrawUserPrimitives( PrimitiveType.TriangleStrip, 2, this.cvTransformedColoredVertexies );
		}
		public void t物理2D描画( Device device, Point pt )
		{
			this.t物理2D描画( device, pt.X, pt.Y, 1f, new Rectangle( 0, 0, this.sz画像サイズ.Width, this.sz画像サイズ.Height ) );
		}
		public void t物理2D描画( Device device, Point pt, Rectangle rc画像内の描画領域 )
		{
			this.t物理2D描画( device, pt.X, pt.Y, 1f, rc画像内の描画領域 );
		}
		public void t物理2D描画( Device device, Point pt, float depth, Rectangle rc画像内の描画領域 )
		{
			this.t物理2D描画( device, pt.X, pt.Y, depth, rc画像内の描画領域 );
		}


		// その他

		#region [ IDosposable 実装 ]
		//-----------------
		public void Dispose()
		{
			C共通.tDisposeする( this.texture ); this.texture = null;
		}
		//-----------------
		#endregion

		#region [ protected ]
		//-----------------
		protected int _透明度;
		protected PositionColoredTexturedVertex[] cvPositionColoredVertexies = new PositionColoredTexturedVertex[]
		{
			new PositionColoredTexturedVertex(),
			new PositionColoredTexturedVertex(),
			new PositionColoredTexturedVertex(),
			new PositionColoredTexturedVertex(),
		};
		protected TransformedColoredTexturedVertex[] cvTransformedColoredVertexies = new TransformedColoredTexturedVertex[]
		{
			new TransformedColoredTexturedVertex(),
			new TransformedColoredTexturedVertex(),
			new TransformedColoredTexturedVertex(),
			new TransformedColoredTexturedVertex(),
		};

		protected void tレンダリングステートの設定( Device device )
		{
			if( this.b加算合成 )
			{
				device.SetRenderState( RenderState.AlphaBlendEnable, true );
				device.SetRenderState( RenderState.SourceBlend, 5 );
				device.SetRenderState( RenderState.DestinationBlend, 2 );
			}
			else
			{
				device.SetRenderState( RenderState.AlphaBlendEnable, true );
				device.SetRenderState( RenderState.SourceBlend, 5 );
				device.SetRenderState( RenderState.DestinationBlend, 6 );
			}
		}
		protected Size t指定されたサイズを超えない最適なテクスチャサイズを返す( Device device, Size sz指定サイズ )
		{
			bool bNonPow2Conditional = ( device.Capabilities.TextureCaps & TextureCaps.NonPow2Conditional ) != 0;
			bool bPow2 = ( device.Capabilities.TextureCaps & TextureCaps.Pow2 ) != 0;
			bool bSquareOnly = ( device.Capabilities.TextureCaps & TextureCaps.SquareOnly ) != 0;
			int maxTextureWidth = device.Capabilities.MaxTextureWidth;
			int maxTextureHeight = device.Capabilities.MaxTextureHeight;
			Size size = new Size( sz指定サイズ.Width, sz指定サイズ.Height );

			if( bPow2 && !bNonPow2Conditional )
			{
				int n = 1;
				do
				{
					n *= 2;
				}
				while( n <= sz指定サイズ.Width );
				sz指定サイズ.Width = n;
				
				n = 1;
				do
				{
					n *= 2;
				}
				while( n <= sz指定サイズ.Height );
				sz指定サイズ.Height = n;
			}

			if( sz指定サイズ.Width > maxTextureWidth )
				sz指定サイズ.Width = maxTextureWidth;

			if( sz指定サイズ.Height > maxTextureHeight )
				sz指定サイズ.Height = maxTextureHeight;

			if( bSquareOnly )
			{
				if( size.Width > size.Height )
				{
					size.Height = size.Width;
					return size;
				}
				if( size.Width < size.Height )
				{
					size.Width = size.Height;
				}
			}

			return size;
		}
		//-----------------
		#endregion
	}
}
