﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using System.Drawing;
using System.Drawing.Imaging;
using SlimDX.Direct3D9;

namespace FDK
{
	/// <summary>
	/// <para>画面上に任意の文字列を短時間だけ表示するフリップボード。</para>
	/// <para>t表示() を呼び出すと、画面最上部にその文字列がしばらく表示され、その後フェードアウトする。</para>
	/// </summary>
	public class CMActフリップボード : CActivity
	{
		public int n表示し続ける時間ms = 1000;
		public int nフェードアウトにかける時間ms = 1000;

		/// <summary>
		/// <para>進行、GUIスレッドから呼び出しても大丈夫。</para>
		/// </summary>
		public void t表示( string str表示文字列 )
		{
			this.ctアニメ用 = new CCounter( 0, this.n表示し続ける時間ms + this.nフェードアウトにかける時間ms, 1 );
			this.str表示文字列 = str表示文字列;
			this.b文字列変更 = true;
		}

		public CMActフリップボード( Format TextureFormat, int n表示し続ける時間ms, int nフェードアウトにかける時間ms )
		{
			this.TextureFormat = TextureFormat;
			this.n表示し続ける時間ms = n表示し続ける時間ms;
			this.nフェードアウトにかける時間ms = nフェードアウトにかける時間ms;
		}
		public CMActフリップボード( Format TextureFormat )
		{
			this.TextureFormat = TextureFormat;
		}
		
		public override void OnManagedリソースの作成( Device D3D9Device )
		{
			if( this.b活性化してない )
				return;

			this.font文字列 = new System.Drawing.Font( FontFamily.GenericSansSerif, 32.0f, FontStyle.Regular, GraphicsUnit.Pixel );
			this.tフリップボードの作成( D3D9Device );

			base.OnManagedリソースの作成( D3D9Device );
		}
		public override void OnManagedリソースの解放()
		{
			if( this.b活性化してない )
				return;

			C共通.tDisposeする( ref this.txフリップボード );
			C共通.tDisposeする( ref this.font文字列 );
			C共通.tDisposeする( ref this.ctアニメ用 );

			base.OnManagedリソースの解放();
		}
		public enum E進行の戻り値 : int { 継続 };
		public override int On進行()
		{
			if( this.b活性化してない ||
				this.ctアニメ用 == null )
				return (int) E進行の戻り値.継続;

			if( this.ctアニメ用.t進行().b終了値に達した )
				this.ctアニメ用 = null;
			
			return (int) E進行の戻り値.継続;
		}
		public override void On描画( Device D3D9Device )
		{
			if( this.b活性化してない ||
				this.txフリップボード == null ||
				this.ctアニメ用 == null )
				return;

			// フリップボードの作成・変更。

			if( this.b文字列変更 )
			{
				this.tフリップボードの作成( D3D9Device, str表示文字列 );
				this.b文字列変更 = false;
			}


			// 透明度の設定。

			if( this.ctアニメ用.n現在の値 < this.n表示し続ける時間ms )
				this.txフリップボード.n透明度 = 255;
			else
				this.txフリップボード.n透明度 = (int) ( ( 1.0 - ( this.ctアニメ用.n現在の値 - this.n表示し続ける時間ms ) / (double) this.nフェードアウトにかける時間ms ) * 255.0 );


			// 描画。

			this.txフリップボード.t2D描画( D3D9Device, 0, 0 );
		}

		#region [ protected ]
		//-----------------
		protected bool b文字列変更 = false;
		protected Format TextureFormat = Format.Unknown;
		protected CTexture txフリップボード = null;
		protected string str表示文字列 = null;
		protected System.Drawing.Font font文字列 = null;		// SlimDX.Direct3D9.Font と重複する
		protected CCounter ctアニメ用 = null;

		protected void tフリップボードの作成( Device D3D9Device, string str表示文字列 )
		{
			this.str表示文字列 = str表示文字列;
			this.tフリップボードの作成( D3D9Device );
		}
		protected void tフリップボードの作成( Device D3D9Device )
		{
			if( string.IsNullOrEmpty( this.str表示文字列 ) ||
				this.font文字列 == null )
				return;

			C共通.tDisposeする( ref this.txフリップボード );

			SizeF sz文字列のサイズ;

			#region [ 文字列のサイズ(SizeF)[px] を取得する。]
			//-----------------
			using( var bitmap = new Bitmap( 1, 1 ) )
			using( Graphics g = Graphics.FromImage( bitmap ) )
			{
				g.PageUnit = GraphicsUnit.Pixel;
				sz文字列のサイズ = g.MeasureString( this.str表示文字列, this.font文字列 );
			}
			//-----------------
			#endregion
			#region [ 取得したサイズで文字列を描画し、フリップボードテクスチャを作成する。 ]
			//-----------------
			using( var bitmap = new Bitmap( (int) sz文字列のサイズ.Width, (int) sz文字列のサイズ.Height, PixelFormat.Format32bppArgb ) )
			using( var g = Graphics.FromImage( bitmap ) )
			{
				// 白地に黒の文字。

				g.Clear( Color.White );
				g.DrawString( this.str表示文字列, this.font文字列, Brushes.Black, 0.0f, 0.0f );
				try
				{
					this.txフリップボード = new CTexture( D3D9Device, bitmap, this.TextureFormat );
				}
				catch
				{
					this.txフリップボード = null;
				}
			}
			//-----------------
			#endregion
		}
		//-----------------
		#endregion
	}
}
