﻿using System;
using System.Collections.Generic;
using System.Text;
using System.Diagnostics;
using System.Drawing;
using SlimDX.Direct3D9;

namespace FDK
{
	/// <summary>
	/// <para>数値＋単位を描画する。</para>
	/// <para>数値が増加すると数字と単位がジャンプする。</para>
	/// <para>本クラスでは、与えられたテクスチャのライフサイクル管理は行わない。</para>
	/// </summary>
	public class CMAct単位付き数値・ジャンプ: CActivity
	{
		/// <summary>
		/// <para>表示する数値。新しい数値をここに上書きしていく。</para>
		/// </summary>
		public int n数値 = 0;

		/// <summary>
		/// <para>false にすると非表示。</para>
		/// </summary>
		public bool b表示 = true;

		public CMAct単位付き数値・ジャンプ()
		{
			// 180度分のジャンプY座標差分を算出。(i:{0 → 90 → 179} で、[i]:{0 → -15 → 0})
			
			for( int i = 0; i < 180; i++ )
				this.nジャンプ差分値[ i ] = (int) ( -15.0 * Math.Sin( ( Math.PI * i ) / 180.0 ) );
		}

		public override void On活性化( Device D3D9Device )
		{
			if( this.b活性化してる )
				return;

			this.n数値 = 0;
			this.n現在の数値 = 0;
			this.nジャンプインデックス値 = 99999;
			this.mジャンプインデックス進行 = new C一定間隔処理();

			base.On活性化( D3D9Device );
		}
		public override void On非活性化()
		{
			if( this.b活性化してない )
				return;

			C共通.tDisposeする( ref this.mジャンプインデックス進行 );

			base.On非活性化();
		}
		public override int On進行()
		{
			if( this.b活性化してない )
				return 0;

			#region [ nジャンプインデックス値 の進行。]
			//-----------------
			if( this.nジャンプインデックス値 < 360 )
			{
				this.mジャンプインデックス進行.t進行( 2, 
					() => {
						if( this.nジャンプインデックス値 < 2000 )
							this.nジャンプインデックス値 += 3;
					} );
			}
			//-----------------
			#endregion

			#region [ 数値が増加していればジャンプを開始する。]
			//-----------------
			if( this.n現在の数値 < n数値 )
			{
				this.nジャンプインデックス値 = 0;

				C共通.tDisposeする( ref this.mジャンプインデックス進行 );
				this.mジャンプインデックス進行 = new C一定間隔処理();
			}
			//-----------------
			#endregion

			this.n現在の数値 = this.n数値;

			return 0;
		}
		/// <summary>
		/// <para>tx数値と単位 には、'0'～'9',単位の11種の画像を含み、CTexture.list描画範囲 に11種の矩形が順番に格納されたテクスチャを指定する。</para>
		/// </summary>
		public void On描画( Device D3D9Device, CTexture tx数値と単位, int n中央X, int n下辺Y, int n文字間隔px )
		{
			// 準正常チェック。

			if( this.b活性化してない ||
				tx数値と単位 == null )
				return;

			// 正常系チェック。

			if( !this.b表示 )
				return;


			// 数値と単位を表示。

			int[] n位の数 = new int[ 10 ];	// 表示は10桁もあれば足りるだろう
			int n桁数 = 0;

			#region [ n現在の数値 を桁数ごとに n位の数[] に格納する。（例：n現在の数値=125 のとき n位の数 = { 5,2,1,0,0,0,0,0,0,0 } ） ]
			//-----------------
			int n = this.n現在の数値;
			while( ( n > 0 ) && ( n桁数 < 10 ) )
			{
				n位の数[ n桁数 ] = n % 10;		// 1の位を格納
				n = ( n - ( n % 10 ) ) / 10;	// 右へシフト（例: 12345 → 1234 ）
				n桁数++;
			}
			//-----------------
			#endregion

			#region [ rc単位, n数字と単位を合わせた画像の全幅px を算出。 ]
			//-----------------
			var rc単位 = tx数値と単位.list描画範囲[ 10 ];

			int n数字と単位を合わせた画像の全幅px = rc単位.Width;
			for( int i = 0; i < n桁数; i++ )
				n数字と単位を合わせた画像の全幅px += tx数値と単位.list描画範囲[ n位の数[ i ] ].Width;
			//-----------------
			#endregion
	
			const int n1桁ごとのジャンプの遅れ = 50;	// 1桁につき 50 インデックス遅れる

			#region [ n位の数[] を、"COMBO" → 1の位 → 10の位 … の順に、右から左へ向かって順番に表示する。]
			//-----------------
			int x = n中央X + ( n数字と単位を合わせた画像の全幅px / 2 );		// 右端X
			int y = n下辺Y;
			int nJump = this.nジャンプインデックス値 - ( n桁数 * n1桁ごとのジャンプの遅れ );


			// 単位 を表示。

			x -= rc単位.Width;
			tx数値と単位.t2D描画( D3D9Device, x, n下辺Y - rc単位.Height, rc単位 );


			// 数値を1の位から順に表示。

			for( int i = 0; i < n桁数; i++ )
			{
				var rc = tx数値と単位.list描画範囲[ n位の数[ i ] ];

				x -= rc.Width + n文字間隔px;
				y = n下辺Y - rc.Height;

				nJump = this.nジャンプインデックス値 - ( ( ( n桁数 - i ) - 1 ) * n1桁ごとのジャンプの遅れ );
				if( ( nJump >= 0 ) && ( nJump < 180 ) )
					y += this.nジャンプ差分値[ nJump ];

				tx数値と単位.t2D描画( D3D9Device, x, y, rc );
			}
			//-----------------
			#endregion
		}

		#region [ protected ]
		//-----------------
		protected int n現在の数値 = -1;
		protected int nジャンプインデックス値 = 9999;
		protected int[] nジャンプ差分値 = new int[ 180 ];
		protected C一定間隔処理 mジャンプインデックス進行 = null;
		//-----------------
		#endregion
	}
}
