﻿using System;
using System.Collections.Generic;
using System.Runtime.InteropServices;
using System.Diagnostics;
using SlimDX;
using SlimDX.Direct3D9;

namespace FDK
{
	/// <summary>
	/// <para>一枚画を中心点から拡大・回転・透明化しつつ描画する点フラッシュ。</para>
	/// </summary>
	public class CMAct点フラッシュ・拡大波 : CActivity
	{
		public void t3D発火( Vector3 v3中心, Vector3 v3法線, float f最小倍率, float f最大倍率, int n所要時間ms, float f最大不透明度0to1 )
		{
			for( int i = 0; i < n波の最大数; i++ )
			{
				if( !this.st波[ i ].b使用中 )
				{
					this.st波[ i ].b使用中 = true;
					this.st波[ i ].f最小倍率 = f最小倍率;
					this.st波[ i ].f最大倍率 = f最大倍率;
					this.st波[ i ].n最大不透明度0to255 = (int) ( 255 * f最大不透明度0to1 );
					this.st波[ i ].n所要時間ms = n所要時間ms;
					this.st波[ i ].v3中心 = v3中心;
					this.st波[ i ].v3法線 = v3法線;
					this.st波[ i ].ct進行 = new CCounter( 0, n所要時間ms, 1 );
					break;
				}
			}
		}

		public CMAct点フラッシュ・拡大波( int nフラッシュの最大同時発火数 )
		{
			this.n波の最大数 = nフラッシュの最大同時発火数;
			this.st波 = new ST波[ this.n波の最大数 ];
		}

		public override void On活性化( Device D3D9Device )
		{
			if( this.b活性化してる )
				return;

			// st波[] の内容をリセット。

			for( int i = 0; i < this.n波の最大数; i++ )
			{
				this.st波[ i ] = new ST波() {
					b使用中 = false,
					ct進行 = new CCounter(),
				};
			}

			base.On活性化( D3D9Device );
		}
		public override void On非活性化()
		{
			if( this.b活性化してない )
				return;

			for( int i = 0; i < this.n波の最大数; i++ )
				C共通.tDisposeする( ref this.st波[ i ].ct進行 );

			base.On非活性化();
		}
		public override int On進行()
		{
			if( this.b活性化してない )
				return 0;

			for( int i = 0; i < n波の最大数; i++ )
			{
				if( this.st波[ i ].b使用中 )
				{
					if( this.st波[ i ].ct進行.t進行().b終了値に達した )
					{
						this.st波[ i ].ct進行.t停止();
						this.st波[ i ].b使用中 = false;
					}
				}
			}

			return 0;
		}
		/// <summary>
		/// <para>描画範囲は tx波.list描画範囲[0] とするが、リストが空なら画像全体とする。</para>
		/// </summary>
		public void On描画( Device D3D9Device, CTexture tx波 )
		{
			if( this.b活性化してない ||
				tx波 == null )
				return;

			for( int i = 0; i < n波の最大数; i++ )
			{
				if( !this.st波[ i ].b使用中 )
					continue;

				float f進行割合0to1 = (float) this.st波[ i ].ct進行.n現在の値 / (float) this.st波[ i ].n所要時間ms;


				var mat = Matrix.Identity;

				float f拡大率 = this.st波[ i ].f最小倍率 + ( this.st波[ i ].f最大倍率 - this.st波[ i ].f最小倍率 ) * f進行割合0to1;
				mat *= Matrix.Scaling(
						f拡大率 * CTexture.f画面比率,
						f拡大率 * CTexture.f画面比率,
						1.0f * CTexture.f画面比率 );

				Vector3 v3変換前法線 = new Vector3( 0f, 0f, -1f );
				Vector3 v3変換後法線 = this.st波[ i ].v3法線;
				Vector3 v3変換前後の2法線の回転軸 = Vector3.Cross( v3変換前法線, v3変換後法線 );
				var f法線の回転角度0toPI = (float) Math.Acos( Vector3.Dot( v3変換前法線, v3変換後法線 ) / ( v3変換前法線.Length() * v3変換後法線.Length() ) );
				mat *= Matrix.RotationQuaternion( Quaternion.RotationAxis( v3変換前後の2法線の回転軸, f法線の回転角度0toPI ) );

				mat *= Matrix.Translation( this.st波[ i ].v3中心 );

				tx波.n透明度 = ( f進行割合0to1 < 0.75f ) ? this.st波[ i ].n最大不透明度0to255 : (int) ( ( this.st波[ i ].n最大不透明度0to255 * ( 1.0f - f進行割合0to1 ) ) / 0.25f );

				if( tx波.list描画範囲.Count == 0 )
					tx波.t3D描画( D3D9Device, mat );
				else
					tx波.t3D描画( D3D9Device, mat, tx波.list描画範囲[ 0 ] );
			}
		}

		#region [ protected ]
		//-----------------
		protected struct ST波
		{
			public bool b使用中;
			public float f最小倍率;
			public float f最大倍率;
			public int n最大不透明度0to255;
			public int n所要時間ms;
			public Vector3 v3中心;
			public Vector3 v3法線;
			public CCounter ct進行;
		}
		protected ST波[] st波;
		protected int n波の最大数;
		//-----------------
		#endregion
	}
}
